// Copyright 2020 gorse Project Authors
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
package cache

import (
	"context"
	"io"
	"math"
	"math/rand"
	"os"
	"strconv"
	"testing"
	"time"

	"github.com/fxtlabs/primes"
	"github.com/juju/errors"
	"github.com/samber/lo"
	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/suite"
	"google.golang.org/protobuf/proto"
)

type baseTestSuite struct {
	suite.Suite
	Database
}

func (suite *baseTestSuite) TearDownSuite() {
	err := suite.Database.Close()
	suite.NoError(err)
}

func (suite *baseTestSuite) SetupTest() {
	err := suite.Database.Ping()
	suite.NoError(err)
	err = suite.Database.Purge()
	suite.NoError(err)
}

func (suite *baseTestSuite) TearDownTest() {
	err := suite.Database.Purge()
	suite.NoError(err)
}

func (suite *baseTestSuite) TestInit() {
	err := suite.Database.Init()
	suite.NoError(err)
}

func (suite *baseTestSuite) TestMeta() {
	ctx := context.Background()
	// Set meta string
	err := suite.Database.Set(ctx, String(Key("meta", "1"), "2"), String(Key("meta", "1000"), "10"))
	suite.NoError(err)
	// Get meta string
	value, err := suite.Database.Get(ctx, Key("meta", "1")).String()
	suite.NoError(err)
	suite.Equal("2", value)
	value, err = suite.Database.Get(ctx, Key("meta", "1000")).String()
	suite.NoError(err)
	suite.Equal("10", value)
	// Delete string
	err = suite.Database.Delete(ctx, Key("meta", "1"))
	suite.NoError(err)
	// Get meta not existed
	value, err = suite.Database.Get(ctx, Key("meta", "1")).String()
	suite.True(errors.Is(err, errors.NotFound), err)
	suite.Equal("", value)
	// Set meta int
	err = suite.Database.Set(ctx, Integer(Key("meta", "1"), 2))
	suite.NoError(err)
	// Get meta int
	valInt, err := suite.Database.Get(ctx, Key("meta", "1")).Integer()
	suite.NoError(err)
	suite.Equal(2, valInt)
	// set meta time
	err = suite.Database.Set(ctx, Time(Key("meta", "1"), time.Date(1996, 4, 8, 0, 0, 0, 0, time.UTC)))
	suite.NoError(err)
	// get meta time
	valTime, err := suite.Database.Get(ctx, Key("meta", "1")).Time()
	suite.NoError(err)
	suite.Equal(1996, valTime.Year())
	suite.Equal(time.Month(4), valTime.Month())
	suite.Equal(8, valTime.Day())

	// test set empty
	err = suite.Database.Set(ctx)
	suite.NoError(err)
	// test set duplicate
	err = suite.Database.Set(ctx, String("100", "1"), String("100", "2"))
	suite.NoError(err)
}

func (suite *baseTestSuite) TestSet() {
	ctx := context.Background()
	err := suite.Database.SetSet(ctx, "set", "1")
	suite.NoError(err)
	// test add
	err = suite.Database.AddSet(ctx, "set", "2")
	suite.NoError(err)
	var members []string
	members, err = suite.Database.GetSet(ctx, "set")
	suite.NoError(err)
	suite.Equal([]string{"1", "2"}, members)
	// test rem
	err = suite.Database.RemSet(ctx, "set", "1")
	suite.NoError(err)
	members, err = suite.Database.GetSet(ctx, "set")
	suite.NoError(err)
	suite.Equal([]string{"2"}, members)
	// test set
	err = suite.Database.SetSet(ctx, "set", "3")
	suite.NoError(err)
	members, err = suite.Database.GetSet(ctx, "set")
	suite.NoError(err)
	suite.Equal([]string{"3"}, members)

	// test add empty
	err = suite.Database.AddSet(ctx, "set")
	suite.NoError(err)
	// test set empty
	err = suite.Database.SetSet(ctx, "set")
	suite.NoError(err)
	// test get empty
	members, err = suite.Database.GetSet(ctx, "unknown_set")
	suite.NoError(err)
	suite.Empty(members)
	// test rem empty
	err = suite.Database.RemSet(ctx, "set")
	suite.NoError(err)
}

func (suite *baseTestSuite) TestScan() {
	ctx := context.Background()
	err := suite.Database.Set(ctx, String("1", "1"))
	suite.NoError(err)
	err = suite.Database.SetSet(ctx, "2", "21", "22", "23")
	suite.NoError(err)

	var keys []string
	err = suite.Database.Scan(func(s string) error {
		keys = append(keys, s)
		return nil
	})
	suite.NoError(err)
	suite.ElementsMatch([]string{"1", "2"}, keys)
}

func (suite *baseTestSuite) TestPurge() {
	ctx := context.Background()
	// insert data
	err := suite.Database.Set(ctx, String("key", "value"))
	suite.NoError(err)
	ret := suite.Database.Get(ctx, "key")
	suite.NoError(ret.err)
	suite.Equal("value", ret.value)

	err = suite.Database.AddSet(ctx, "set", "a", "b", "c")
	suite.NoError(err)
	s, err := suite.Database.GetSet(ctx, "set")
	suite.NoError(err)
	suite.ElementsMatch([]string{"a", "b", "c"}, s)

	// purge data
	err = suite.Database.Purge()
	suite.NoError(err)
	ret = suite.Database.Get(ctx, "key")
	suite.ErrorIs(ret.err, errors.NotFound)
	s, err = suite.Database.GetSet(ctx, "set")
	suite.NoError(err)
	suite.Empty(s)

	// purge empty dataset
	err = suite.Database.Purge()
	suite.NoError(err)
}

func (suite *baseTestSuite) TestPushPop() {
	ctx := context.Background()
	err := suite.Push(ctx, "a", "1")
	suite.NoError(err)
	err = suite.Push(ctx, "a", "2")
	suite.NoError(err)
	count, err := suite.Remain(ctx, "a")
	suite.NoError(err)
	suite.Equal(int64(2), count)

	err = suite.Push(ctx, "b", "1")
	suite.NoError(err)
	err = suite.Push(ctx, "b", "2")
	suite.NoError(err)
	err = suite.Push(ctx, "b", "1")
	suite.NoError(err)
	count, err = suite.Remain(ctx, "b")
	suite.NoError(err)
	suite.Equal(int64(2), count)

	value, err := suite.Pop(ctx, "a")
	suite.NoError(err)
	suite.Equal("1", value)
	value, err = suite.Pop(ctx, "a")
	suite.NoError(err)
	suite.Equal("2", value)
	_, err = suite.Pop(ctx, "a")
	suite.ErrorIs(err, io.EOF)

	value, err = suite.Pop(ctx, "b")
	suite.NoError(err)
	suite.Equal("2", value)
	value, err = suite.Pop(ctx, "b")
	suite.NoError(err)
	suite.Equal("1", value)
	_, err = suite.Pop(ctx, "b")
	suite.ErrorIs(err, io.EOF)
}

func (suite *baseTestSuite) TestDocument() {
	ts := time.Date(2023, 1, 1, 0, 0, 0, 0, time.UTC)
	ctx := context.Background()
	err := suite.AddDocuments(ctx, "a", "", []Document{{
		Id:         "0",
		Score:      math.MaxFloat64,
		IsHidden:   true,
		Categories: []string{"a", "b"},
		Timestamp:  ts,
	}})
	suite.NoError(err)
	err = suite.AddDocuments(ctx, "a", "", []Document{{
		Id:         "1",
		Score:      100,
		Categories: []string{"a", "b"},
		Timestamp:  ts,
	}})
	suite.NoError(err)
	err = suite.AddDocuments(ctx, "a", "", []Document{
		{
			Id:         "1",
			Score:      1,
			Categories: []string{"a", "b"},
			Timestamp:  ts,
		},
		{
			Id:         "2",
			Score:      2,
			Categories: []string{"b", "c"},
			Timestamp:  ts,
		},
		{
			Id:         "3",
			Score:      3,
			Categories: []string{"b"},
			Timestamp:  time.Date(2020, 1, 1, 0, 0, 0, 0, time.UTC),
		},
		{
			Id:         "4",
			Score:      4,
			Categories: []string{""},
			Timestamp:  ts,
		},
		{
			Id:         "5",
			Score:      5,
			Categories: []string{"b"},
			Timestamp:  ts,
		},
	})
	suite.NoError(err)
	err = suite.AddDocuments(ctx, "b", "", []Document{{
		Id:         "6",
		Score:      6,
		Categories: []string{"b"},
		Timestamp:  ts,
	}})
	suite.NoError(err)

	// search documents
	documents, err := suite.SearchDocuments(ctx, "a", "", []string{"b"}, 1, 3)
	suite.NoError(err)
	suite.Equal([]Document{
		{Id: "3", Score: 3, Categories: []string{"b"}, Timestamp: time.Date(2020, 1, 1, 0, 0, 0, 0, time.UTC)},
		{Id: "2", Score: 2, Categories: []string{"b", "c"}, Timestamp: ts},
	}, documents)
	documents, err = suite.SearchDocuments(ctx, "a", "", []string{"b"}, 0, -1)
	suite.NoError(err)
	suite.Equal([]Document{
		{Id: "5", Score: 5, Categories: []string{"b"}, Timestamp: ts},
		{Id: "3", Score: 3, Categories: []string{"b"}, Timestamp: time.Date(2020, 1, 1, 0, 0, 0, 0, time.UTC)},
		{Id: "2", Score: 2, Categories: []string{"b", "c"}, Timestamp: ts},
		{Id: "1", Score: 1, Categories: []string{"a", "b"}, Timestamp: ts},
	}, documents)

	// search documents with empty category
	documents, err = suite.SearchDocuments(ctx, "a", "", []string{""}, 0, -1)
	suite.NoError(err)
	suite.Equal([]Document{{Id: "4", Score: 4, Categories: []string{""}, Timestamp: ts}}, documents)

	// delete nothing
	err = suite.DeleteDocuments(ctx, []string{"a"}, DocumentCondition{})
	suite.ErrorIs(err, errors.NotValid)
	// delete by value
	err = suite.DeleteDocuments(ctx, []string{"a"}, DocumentCondition{Id: proto.String("5")})
	suite.NoError(err)
	documents, err = suite.SearchDocuments(ctx, "a", "", []string{"b"}, 0, 1)
	suite.NoError(err)
	suite.Len(documents, 1)
	suite.Equal("3", documents[0].Id)
	// delete by timestamp
	err = suite.DeleteDocuments(ctx, []string{"a"}, DocumentCondition{Before: lo.ToPtr(time.Date(2022, 1, 1, 0, 0, 0, 0, time.UTC))})
	suite.NoError(err)
	documents, err = suite.SearchDocuments(ctx, "a", "", []string{"b"}, 0, 1)
	suite.NoError(err)
	suite.Len(documents, 1)
	suite.Equal("2", documents[0].Id)

	// update categories
	err = suite.UpdateDocuments(ctx, []string{"a"}, "2", DocumentPatch{Categories: []string{"c", "s"}})
	suite.NoError(err)
	documents, err = suite.SearchDocuments(ctx, "a", "", []string{"s"}, 0, 1)
	suite.NoError(err)
	suite.Len(documents, 1)
	suite.Equal("2", documents[0].Id)
	err = suite.UpdateDocuments(ctx, []string{"a"}, "2", DocumentPatch{Categories: []string{"c"}})
	suite.NoError(err)
	documents, err = suite.SearchDocuments(ctx, "a", "", []string{"s"}, 0, 1)
	suite.NoError(err)
	suite.Empty(documents)

	// update is hidden
	err = suite.UpdateDocuments(ctx, []string{"a"}, "0", DocumentPatch{IsHidden: proto.Bool(false)})
	suite.NoError(err)
	documents, err = suite.SearchDocuments(ctx, "a", "", []string{"b"}, 0, 1)
	suite.NoError(err)
	suite.Len(documents, 1)
	suite.Equal("0", documents[0].Id)
}

func (suite *baseTestSuite) TestSubsetDocument() {
	ts := time.Date(2023, 1, 1, 0, 0, 0, 0, time.UTC)
	ctx := context.Background()
	err := suite.AddDocuments(ctx, "a", "a", []Document{
		{
			Id:         "1",
			Score:      1,
			Categories: []string{"a", "b"},
			Timestamp:  ts,
		},
		{
			Id:         "2",
			Score:      2,
			Categories: []string{"b", "c"},
			Timestamp:  ts,
		},
		{
			Id:         "3",
			Score:      3,
			Categories: []string{"b"},
			Timestamp:  ts,
		},
	})
	suite.NoError(err)
	err = suite.AddDocuments(ctx, "b", "", []Document{
		{
			Id:         "4",
			Score:      4,
			Categories: []string{},
			Timestamp:  ts,
		},
		{
			Id:         "3",
			Score:      3,
			Categories: []string{"b"},
			Timestamp:  ts,
		},
		{
			Id:         "2",
			Score:      2,
			Categories: []string{"b"},
			Timestamp:  ts,
		},
	})
	suite.NoError(err)

	// search documents
	documents, err := suite.SearchDocuments(ctx, "a", "a", []string{"b"}, 0, -1)
	suite.NoError(err)
	suite.Equal([]Document{
		{Id: "3", Score: 3, Categories: []string{"b"}, Timestamp: ts},
		{Id: "2", Score: 2, Categories: []string{"b", "c"}, Timestamp: ts},
		{Id: "1", Score: 1, Categories: []string{"a", "b"}, Timestamp: ts},
	}, documents)

	// update categories
	err = suite.UpdateDocuments(ctx, []string{"a", "b"}, "2", DocumentPatch{Categories: []string{"b", "s"}})
	suite.NoError(err)
	documents, err = suite.SearchDocuments(ctx, "a", "a", []string{"s"}, 0, 1)
	suite.NoError(err)
	suite.Len(documents, 1)
	suite.Equal("2", documents[0].Id)
	documents, err = suite.SearchDocuments(ctx, "b", "", []string{"s"}, 0, 1)
	suite.NoError(err)
	suite.Len(documents, 1)
	suite.Equal("2", documents[0].Id)

	// delete by value
	err = suite.DeleteDocuments(ctx, []string{"a", "b"}, DocumentCondition{Id: proto.String("3")})
	suite.NoError(err)
	documents, err = suite.SearchDocuments(ctx, "a", "a", []string{"b"}, 0, 1)
	suite.NoError(err)
	suite.Len(documents, 1)
	suite.Equal("2", documents[0].Id)
	documents, err = suite.SearchDocuments(ctx, "b", "", []string{"b"}, 0, 1)
	suite.NoError(err)
	suite.Len(documents, 1)
	suite.Equal("2", documents[0].Id)

	// delete in subset
	err = suite.DeleteDocuments(ctx, []string{"a", "b"}, DocumentCondition{
		Subset: proto.String("a"),
		Id:     proto.String("2"),
	})
	suite.NoError(err)
	documents, err = suite.SearchDocuments(ctx, "a", "a", []string{"b"}, 0, 1)
	suite.NoError(err)
	suite.Len(documents, 1)
	suite.Equal("1", documents[0].Id)
	documents, err = suite.SearchDocuments(ctx, "b", "", []string{"b"}, 0, 1)
	suite.NoError(err)
	suite.Len(documents, 1)
	suite.Equal("2", documents[0].Id)
}

func (suite *baseTestSuite) TestTimeSeries() {
	ts := time.Date(2023, 1, 1, 0, 0, 0, 0, time.UTC)
	ctx := context.Background()
	err := suite.AddTimeSeriesPoints(ctx, []TimeSeriesPoint{
		{Name: "a", Value: 1, Timestamp: ts.Add(1 * time.Second)},
		{Name: "a", Value: 2, Timestamp: ts.Add(2 * time.Second)},
		{Name: "a", Value: 3, Timestamp: ts.Add(3 * time.Second)},
		{Name: "a", Value: 4, Timestamp: ts.Add(4 * time.Second)},
		{Name: "a", Value: 5, Timestamp: ts.Add(5 * time.Second)}})
	suite.NoError(err)

	points, err := suite.GetTimeSeriesPoints(ctx, "a", ts.Add(2*time.Second), ts.Add(5*time.Second))
	suite.NoError(err)
	suite.Equal([]TimeSeriesPoint{
		{Name: "a", Value: 2, Timestamp: ts.Add(2 * time.Second)},
		{Name: "a", Value: 3, Timestamp: ts.Add(3 * time.Second)},
		{Name: "a", Value: 4, Timestamp: ts.Add(4 * time.Second)},
	}, points)
}

func TestKey(t *testing.T) {
	assert.Empty(t, Key())
	assert.Equal(t, "a", Key("a"))
	assert.Equal(t, "a", Key("a", ""))
	assert.Equal(t, "a/b", Key("a", "b"))
}

var (
	benchmarkDataSize = 100000
	primeTable        []int
)

func init() {
	benchmarkDataSizeStr := os.Getenv("BENCHMARK_DATA_SIZE")
	if benchmarkDataSizeStr != "" {
		benchmarkDataSize, _ = strconv.Atoi(benchmarkDataSizeStr)
	}
	primeTable = primes.Sieve(benchmarkDataSize)
}

func primeFactor(n int) []int {
	var factors []int
	for _, p := range primeTable {
		if n%p == 0 {
			factors = append(factors, p)
		}
	}
	return factors
}

func benchmark(b *testing.B, database Database) {
	b.Run("AddDocuments", func(b *testing.B) {
		benchmarkAddDocuments(b, database)
	})
	b.Run("SearchDocuments", func(b *testing.B) {
		benchmarkSearchDocuments(b, database)
	})
	b.Run("UpdateDocuments", func(b *testing.B) {
		benchmarkUpdateDocuments(b, database)
	})
}

func benchmarkAddDocuments(b *testing.B, database Database) {
	ctx := context.Background()
	var documents []Document
	for i := 1; i <= b.N; i++ {
		documents = append(documents, Document{
			Id:         strconv.Itoa(i),
			Score:      float64(-i),
			Categories: lo.Map(primeFactor(i), func(n, _ int) string { return strconv.Itoa(n) }),
			Timestamp:  time.Date(2023, 1, 1, 0, 0, 0, 0, time.UTC),
		})
	}
	b.ResetTimer()
	err := database.AddDocuments(ctx, "a", "", documents)
	assert.NoError(b, err)
}

func benchmarkSearchDocuments(b *testing.B, database Database) {
	// insert data
	ctx := context.Background()
	var documents []Document
	for i := 1; i <= benchmarkDataSize; i++ {
		documents = append(documents, Document{
			Id:         strconv.Itoa(i),
			Score:      float64(-i),
			Categories: lo.Map(primeFactor(i), func(n, _ int) string { return strconv.Itoa(n) }),
			Timestamp:  time.Date(2023, 1, 1, 0, 0, 0, 0, time.UTC),
		})
	}
	err := database.AddDocuments(ctx, "a", "", documents)
	assert.NoError(b, err)
	// search data
	b.ResetTimer()
	for i := 0; i < b.N; i++ {
		// select a random prime
		p := primeTable[rand.Intn(len(primeTable))]
		// search documents
		r, err := database.SearchDocuments(ctx, "a", "", []string{strconv.Itoa(p)}, 0, 10)
		assert.NoError(b, err)
		assert.NotEmpty(b, r)
	}
}

func benchmarkUpdateDocuments(b *testing.B, database Database) {
	ctx := context.Background()
	b.ResetTimer()
	for i := 1; i <= b.N; i++ {
		// select a random number
		n := rand.Intn(benchmarkDataSize) + 1
		// update documents
		err := database.UpdateDocuments(ctx, []string{"a"}, strconv.Itoa(n), DocumentPatch{
			Score: proto.Float64(float64(n)),
		})
		assert.NoError(b, err)
	}
}
